﻿/*
Copyright (c) 2014 Stephen Stair (sgstair@akkit.org)
Additional code Miguel Parra (miguelvp@msn.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

// 
// the winusbdotnet repo isn't the correct place for this code long term
// Code is here for now for convenience in testing and iteration while being developed.
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Windows.Forms;
using System.IO;

using winusbdotnet.UsbDevices;

namespace TestSeek
{
    public partial class Form1 : Form
    {

        SeekThermal thermal;
        Thread thermalThread;
        int frameCount;
        bool stopThread;
        bool captureReference;
        bool rangeOverrideToggle;
        bool useAutomaticRange;

        ThermalFrame lastFrame, lastCalibrationFrame, lastReferenceFrame;
        ThermalFrame frameID4, frameID9, frameID8, frameID7, frameID10, frameID5;
        CalibratedThermalFrame lastUsableFrame, lastRenderedFrame;

        Queue<Bitmap> bmpQueue;

        public Form1()
        {
            InitializeComponent();

            DoubleBuffered = true;
            bmpQueue = new Queue<Bitmap>();

            // Init button trigger to be off.
            captureReference = false;
            // Initialize the range override toggle.
            rangeOverrideToggle = false;
            // Use automatic range by default.
            useAutomaticRange = true;

            var device = SeekThermal.Enumerate().FirstOrDefault();
            if(device == null)
            {
                MessageBox.Show("No Seek Thermal devices found.");
                return;
            }
            thermal = new SeekThermal(device);

            thermalThread = new Thread(ThermalThreadProc);
            thermalThread.IsBackground = true;
            thermalThread.Start();
        }

        void ThermalThreadProc()
        {
            BinaryWriter bw;
            BinaryWriter tw;
            DateTime currentFrameTime = DateTime.Now;
            DateTime previousFrameTime = currentFrameTime;
            DateTime currentTime = DateTime.Now;
            DateTime previousTime = currentTime;
            int framesToCapture = 20;

            // Initialize frame (1 based)
            frameCount = 1;

            // Create the output files to save first 20 frames and associated metadata.
            bw = new BinaryWriter(new FileStream("data.dat", FileMode.Create));
            tw = new BinaryWriter(new FileStream("data.txt", FileMode.Create));

            while (!stopThread && thermal != null)
            {
                bool progress = false;

                // Store the previous frame time and compute the new one.
                previousFrameTime = currentFrameTime;
                currentFrameTime = DateTime.Now;

                // Get frame
                lastFrame = thermal.GetFrameBlocking();

                // Keep the first 6 frames, or anytime those frame IDs are encountered.
                // They might be usefull for image processing.
                switch (lastFrame.RawDataU16[10])
                {
                    case 4:
                        frameID4 = lastFrame;
                        //frameID4.Inverse();
                        break;
                    case 9:
                        frameID9 = lastFrame;
                        break;
                    case 8:
                        frameID8 = lastFrame;
                        break;
                    case 7:
                        frameID7 = lastFrame;
                        break;
                    case 10:
                        frameID10 = lastFrame;
                        break;
                    case 5:
                        frameID5 = lastFrame;
                        break;
                    default:
                        // Ignore the rest from safe keeping, since they are dealt with later.
                        //
                        // ID 1 is a calibration frame
                        // ID 6 is a pre-calibration frame
                        // ID 3 is a visual frame.
                        break;
                }

                // Time after frame capture
                previousTime = currentTime;
                currentTime = DateTime.Now;

                // Save data and metadata for the first framesToCapture frames
                if (frameCount <= framesToCapture)
                {
                    // Write frame in binary file.
                    bw.Write(lastFrame.RawData);

                    // Only add the file write delay if it's not the 1st frame and there was some time elapsed.
                    if ((frameCount > 1) && (currentFrameTime.Ticks > previousTime.Ticks))
                    {
                        tw.Write(Encoding.ASCII.GetBytes(String.Format(" <- {0:F4}ms File Write delay\n\n", (currentFrameTime.Ticks - previousTime.Ticks) / 10000.0)));
                    }
                    // Write frame count, frame ID, start timestamp, end timestamp and milliseconds elapsed time.
                    // Timestamps are 64 bit in hundreds of nanoseconds since Jan 1, 2001 at midnight.
                    tw.Write(Encoding.ASCII.GetBytes(String.Format("Frame {0} ID {1}\n", frameCount, lastFrame.RawDataU16[10])));
                    tw.Write(Encoding.ASCII.GetBytes(String.Format("{0:X}\n", currentFrameTime.Ticks)));
                    tw.Write(Encoding.ASCII.GetBytes(String.Format("{0:X} <- {1:F4}ms\n\n", currentTime.Ticks, (currentTime.Ticks - currentFrameTime.Ticks)/10000.0)));

                    if (frameCount == framesToCapture)
                    {
                        // Close files after framesToCapture frames.
                        bw.Close();
                        tw.Close();
                    }
                }

                if(lastFrame.IsCalibrationFrame)
                {
                    lastCalibrationFrame = lastFrame;
                }
                else
                {
                    if(lastCalibrationFrame != null && lastFrame.IsUsableFrame)
                    {
                        if (lastReferenceFrame != null && lastReferenceFrame.IsUsableFrame)
                        {
                            lastUsableFrame = lastFrame.ProcessFrameU16(lastReferenceFrame, frameID4);
                        }
                        else
                        {
                            lastUsableFrame = lastFrame.ProcessFrameU16(lastCalibrationFrame, frameID4);
                        }

                        // Toggle bettwen internal or external calibration frame.
                        if (captureReference && (lastFrame.RawDataU16[10] == 3))
                        {
                            if (lastReferenceFrame != null)
                            {
                                lastReferenceFrame = null;
                            }
                            else
                            {
                                lastReferenceFrame = lastFrame;
                            }
                            captureReference = false;
                        }
                        progress = true;
                    }
                }

                // Toggle between automatic to manual range.
                if (rangeOverrideToggle)
                {
                    useAutomaticRange = !useAutomaticRange;
                    rangeOverrideToggle = false;
                }

                // Increase frame count.
                frameCount++;

                if(progress)
                {
                    Invalidate();
                }
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            stopThread = true;
            if (thermal != null)
            {
                thermalThread.Join(500);
                thermal.Deinit();
            }
        }

        public struct palette
        {
            public int r, g, b;
            public palette(int ri, int gi, int bi)
            {
                r = ri;
                g = gi;
                b = bi;
            }
        }

        private void Form1_Paint(object sender, PaintEventArgs e)
        {
            CalibratedThermalFrame data = lastUsableFrame;
            if (data == null) return;
            int y;
            if(data != lastRenderedFrame)
            {
                lastRenderedFrame = data;
                // Process new frame
                Bitmap bmp = new Bitmap((data.Width-2)*2, data.Height*2);
                int c = 0;
                int v;
                ushort maxValue, minValue;

                // Would be better to have a computed palette larger than 256 entries since it's color based it could be the full 14 bits the sensor supports.
                // But for now this 256 iron palette will do.
                palette[] IronPalette = {
                    new palette(000, 000, 000), new palette(000, 000, 007), new palette(000, 000, 024), new palette(000, 000, 038), new palette(000, 000, 046), new palette(000, 000, 052), new palette(000, 000, 059), new palette(000, 000, 066),
                    new palette(000, 000, 073), new palette(000, 000, 080), new palette(001, 000, 085), new palette(002, 000, 089), new palette(003, 000, 093), new palette(004, 000, 097), new palette(005, 000, 101), new palette(006, 000, 104),
                    new palette(008, 000, 108), new palette(010, 000, 112), new palette(011, 000, 116), new palette(013, 000, 118), new palette(015, 000, 119), new palette(017, 000, 121), new palette(020, 000, 123), new palette(024, 000, 126),
                    new palette(027, 000, 128), new palette(030, 000, 130), new palette(033, 000, 133), new palette(036, 000, 135), new palette(040, 000, 136), new palette(043, 000, 138), new palette(047, 000, 139), new palette(050, 000, 141),
                    new palette(053, 000, 142), new palette(056, 000, 144), new palette(059, 000, 145), new palette(062, 000, 146), new palette(064, 000, 148), new palette(066, 000, 149), new palette(069, 000, 150), new palette(072, 000, 150),
                    new palette(075, 000, 150), new palette(078, 000, 151), new palette(081, 000, 151), new palette(083, 000, 152), new palette(087, 000, 152), new palette(090, 000, 153), new palette(093, 000, 154), new palette(096, 000, 155),
                    new palette(099, 000, 155), new palette(102, 000, 155), new palette(105, 000, 155), new palette(108, 000, 156), new palette(111, 000, 156), new palette(113, 000, 157), new palette(115, 000, 157), new palette(119, 000, 157),
                    new palette(122, 000, 157), new palette(125, 000, 157), new palette(128, 000, 157), new palette(130, 000, 157), new palette(133, 000, 157), new palette(136, 000, 157), new palette(138, 000, 157), new palette(141, 000, 157),
                    new palette(144, 000, 156), new palette(147, 000, 156), new palette(150, 000, 155), new palette(153, 000, 155), new palette(155, 000, 155), new palette(157, 000, 155), new palette(160, 000, 155), new palette(162, 000, 155),
                    new palette(164, 000, 155), new palette(167, 000, 154), new palette(169, 000, 154), new palette(171, 000, 153), new palette(172, 000, 153), new palette(174, 001, 152), new palette(176, 001, 152), new palette(177, 001, 151),
                    new palette(179, 001, 151), new palette(181, 002, 150), new palette(183, 002, 149), new palette(184, 003, 149), new palette(185, 004, 149), new palette(187, 005, 148), new palette(188, 005, 147), new palette(190, 005, 146),
                    new palette(191, 006, 146), new palette(192, 007, 145), new palette(193, 008, 144), new palette(194, 009, 143), new palette(195, 011, 142), new palette(196, 012, 141), new palette(197, 013, 140), new palette(199, 014, 138),
                    new palette(200, 016, 136), new palette(201, 018, 134), new palette(202, 019, 133), new palette(203, 020, 131), new palette(205, 022, 129), new palette(206, 023, 127), new palette(207, 025, 124), new palette(208, 026, 121),
                    new palette(209, 027, 119), new palette(210, 029, 116), new palette(211, 031, 114), new palette(212, 033, 112), new palette(213, 034, 108), new palette(214, 036, 104), new palette(215, 038, 101), new palette(216, 040, 098),
                    new palette(217, 042, 095), new palette(218, 045, 092), new palette(219, 047, 087), new palette(220, 048, 082), new palette(221, 049, 077), new palette(222, 051, 071), new palette(223, 053, 065), new palette(224, 055, 060),
                    new palette(224, 056, 054), new palette(225, 058, 049), new palette(226, 060, 043), new palette(227, 062, 037), new palette(228, 063, 032), new palette(228, 065, 028), new palette(229, 067, 026), new palette(230, 069, 024),
                    new palette(230, 071, 021), new palette(231, 073, 019), new palette(232, 075, 017), new palette(232, 076, 015), new palette(233, 077, 013), new palette(234, 078, 012), new palette(235, 080, 011), new palette(235, 082, 010),
                    new palette(235, 084, 009), new palette(236, 086, 008), new palette(236, 088, 008), new palette(237, 090, 007), new palette(237, 091, 006), new palette(238, 092, 005), new palette(238, 094, 005), new palette(239, 095, 004),
                    new palette(239, 097, 004), new palette(240, 099, 003), new palette(240, 101, 003), new palette(241, 102, 003), new palette(241, 103, 003), new palette(241, 105, 002), new palette(241, 106, 002), new palette(241, 108, 001),
                    new palette(242, 109, 001), new palette(242, 111, 001), new palette(243, 113, 001), new palette(243, 114, 001), new palette(244, 116, 000), new palette(244, 117, 000), new palette(244, 119, 000), new palette(244, 121, 000),
                    new palette(245, 124, 000), new palette(245, 126, 000), new palette(246, 128, 000), new palette(246, 130, 000), new palette(247, 131, 000), new palette(247, 133, 000), new palette(248, 135, 000), new palette(248, 136, 000),
                    new palette(248, 137, 000), new palette(248, 139, 000), new palette(248, 140, 000), new palette(249, 142, 000), new palette(249, 143, 000), new palette(249, 144, 000), new palette(249, 146, 000), new palette(250, 148, 000),
                    new palette(250, 150, 000), new palette(251, 152, 000), new palette(251, 155, 000), new palette(252, 157, 000), new palette(252, 159, 000), new palette(253, 161, 000), new palette(253, 163, 000), new palette(253, 166, 000),
                    new palette(253, 168, 000), new palette(253, 170, 000), new palette(253, 172, 000), new palette(253, 174, 000), new palette(254, 175, 000), new palette(254, 177, 000), new palette(254, 178, 000), new palette(254, 180, 000),
                    new palette(254, 183, 000), new palette(254, 185, 000), new palette(254, 186, 000), new palette(254, 188, 000), new palette(254, 189, 000), new palette(254, 191, 000), new palette(254, 193, 000), new palette(254, 195, 000),
                    new palette(254, 197, 000), new palette(254, 199, 000), new palette(254, 200, 000), new palette(254, 202, 001), new palette(254, 203, 001), new palette(254, 204, 002), new palette(254, 206, 003), new palette(254, 207, 004),
                    new palette(254, 209, 006), new palette(254, 211, 008), new palette(254, 213, 010), new palette(254, 215, 011), new palette(254, 216, 012), new palette(255, 218, 014), new palette(255, 219, 016), new palette(255, 220, 019),
                    new palette(255, 221, 023), new palette(255, 222, 027), new palette(255, 224, 031), new palette(255, 225, 035), new palette(255, 227, 038), new palette(255, 228, 042), new palette(255, 229, 048), new palette(255, 230, 053),
                    new palette(255, 231, 060), new palette(255, 233, 065), new palette(255, 234, 071), new palette(255, 235, 077), new palette(255, 237, 083), new palette(255, 238, 089), new palette(255, 239, 096), new palette(255, 239, 102),
                    new palette(255, 240, 109), new palette(255, 241, 115), new palette(255, 241, 124), new palette(255, 242, 132), new palette(255, 243, 139), new palette(255, 244, 146), new palette(255, 244, 153), new palette(255, 245, 160),
                    new palette(255, 245, 168), new palette(255, 246, 175), new palette(255, 247, 181), new palette(255, 248, 187), new palette(255, 248, 193), new palette(255, 249, 198), new palette(255, 249, 204), new palette(255, 250, 210),
                    new palette(255, 251, 216), new palette(255, 252, 222), new palette(255, 253, 227), new palette(255, 253, 232), new palette(255, 254, 237), new palette(255, 254, 243), new palette(255, 255, 247), new palette(255, 255, 249)
                };

                // Display last sensor min/max values.
                label2.Text = "Max: " + lastFrame.MaxValue;
                label3.Text = "Min: " + lastFrame.MinValue;

                // Display last calibration min/max values (internal or external)
                if (lastReferenceFrame != null && lastReferenceFrame.IsUsableFrame)
                {
                    // Update label for calibration button functionality.
                    label1.Text = "Push button to switch to default calibration";
                    button1.Text = "Calibraton: Use Internal";
                    label6.Text = "Max: " + lastReferenceFrame.MaxValue;
                    label7.Text = "Min: " + lastReferenceFrame.MinValue;
                }
                else
                {
                    // Update label for calibration button functionality.
                    label1.Text = "Face camera down on an even heated surface";
                    button1.Text = "Calibraton: Use Frame";
                    label6.Text = "Max: " + lastCalibrationFrame.MaxValue;
                    label7.Text = "Min: " + lastCalibrationFrame.MinValue;
                }

                if (useAutomaticRange)
                {
                    button2.Text = "Manual";
                    maxValue = data.MaxValue;
                    minValue = data.MinValue;
                    trackBar1.Value = maxValue;
                    trackBar2.Value = minValue;
                }
                else
                {
                    button2.Text = "Automatic";
                    maxValue = (ushort)trackBar1.Value;
                    minValue = (ushort)trackBar2.Value;
                }

                // Noise reduction
                //smaller number means less noise but also less details, 25 seems to be a good value
                // 25 was the original comment in frenky's sample code, since this is not the same code I think 4 is a good compromise.
                int noiseReduction = 4 * (data.MaxValue - data.MinValue) / 256; 
                ushort[] arrColor = new ushort[5];
                ushort min, max;
                ulong temp,sum;
                ulong value;

                // Three passes works well, maybe too agressive.
                for (int n = 0; n < 3; n++)
                {
                    for (y = 0; y < 154; y++)
                    {
                        for (int x = 0; x < 206; x++)
                        {
                            min = 0xffff;
                            max = 0;
                            sum = 0;
                            arrColor[0] = data.PixelData[x + 208 * (y + 1)];
                            arrColor[1] = data.PixelData[x + 2 + 208 * (y + 1)];
                            arrColor[2] = data.PixelData[x + 1 + 208 * (y)];
                            arrColor[3] = data.PixelData[x + 1 + 208 * (y + 2)];
                            for (int i = 0; i < 4; i++)
                            {
                                if (arrColor[i] > max) max = arrColor[i];
                                if (arrColor[i] < min) min = arrColor[i];
                                sum += arrColor[i];
                            }

                            //get average value, but exclude neighbour dead pixels from average:
                            temp = min;
                            temp += max;
                            value = sum;
                            value -= temp;
                            value /= 2;

                            arrColor[4] = data.PixelData[x + 1 + 208 * (y + 1)];
                            if (Math.Abs((ushort)value - arrColor[4]) > noiseReduction) //if image is too different use average color from near pixels
                            {
                                data.PixelData[x + 1 + 208 * (y + 1)] = (ushort)value;
                            }
                        }
                    }
                }

                // Clean up the borders so we don't have strange data
                // we loose the original value and we use the interpolation of the pixels next to the edge
                // for all the edges we care about (top bottom and left).
                ulong average;
                y = 0;
                for (int x = 1; x < 206; x++)
                {
                    c = y * 208 + x;
                    average = data.PixelData[c + 207];
                    average += data.PixelData[c + 208];
                    average += data.PixelData[c + 209];
                    average /= 3;
                    data.PixelData[c] = (ushort)average;
                }
                y = 155;
                for (int x = 1; x < 206; x++)
                {
                    c = y * 208 + x;
                    average = data.PixelData[c - 207];
                    average += data.PixelData[c - 208];
                    average += data.PixelData[c - 209];
                    average /= 3;
                    data.PixelData[c] = (ushort)average;
                }
                for (y = 1; y < 155; y++)
                {
                    c = y * 208;
                    average = data.PixelData[c - 207];
                    average += data.PixelData[c + 1];
                    average += data.PixelData[c + 209];
                    average /= 3;
                    data.PixelData[c] = (ushort)average;
                }

                // Recompute the min max after the image is all fixed
                // unless the sliders have overriden values, then it doesn't matter.
                if (useAutomaticRange)
                {
                    ushort val;
                    minValue = 0xFFFF;
                    maxValue = 0x0000;
                    for (y = 0; y < 156; y++)
                    {
                        for (int x = 0; x < 206; x++)
                        {
                            val = data.PixelData[y * 208 + x];
                            if (val < minValue) minValue = val;
                            if (val > maxValue) maxValue = val;
                        }
                    }
                }

                // Display displayed data min/max value.
                label9.Text = "Max: " + maxValue;
                label10.Text = "Min: " + minValue;

                for (y = 0; y < 156; y++)
                {
                    for (int x = 0; x < 206; x++)
                    {
                        v = data.PixelData[y * 208 + x]; // + data.PixelData[y * 208 + 206] / 10; // no need to use column 207 since we already use frame id 4, max/min will be off if uncommented as well.

                        // Scale data to be within [0-255] for LUT mapping.
                        ushort maxmin = maxValue;
                        maxmin -= minValue;
                        // Avoid divide by 0
                        if (maxmin == 0)
                            maxmin = 1;

                        v = (v - minValue) * 256 / maxmin;
                        if (v < 0)
                            v = 0;
                        if (v > 255)
                            v = 255;

                        // Greyscale output (would always be limited to 256 colors)
                        //bmp.SetPixel(x * 2, y * 2, Color.FromArgb(v, v, v));
                        //bmp.SetPixel(x * 2 + 1, y * 2, Color.FromArgb(v, v, v));
                        //bmp.SetPixel(x * 2, y * 2 + 1, Color.FromArgb(v, v, v));
                        //bmp.SetPixel(x * 2 + 1, y * 2 + 1, Color.FromArgb(v, v, v));

                        // Apply lookup table.
                        bmp.SetPixel(x * 2, y * 2, Color.FromArgb(IronPalette[v].r, IronPalette[v].g, IronPalette[v].b));
                        bmp.SetPixel(x * 2 + 1, y * 2, Color.FromArgb(IronPalette[v].r, IronPalette[v].g, IronPalette[v].b));
                        bmp.SetPixel(x * 2, y * 2 + 1, Color.FromArgb(IronPalette[v].r, IronPalette[v].g, IronPalette[v].b));
                        bmp.SetPixel(x * 2 + 1, y * 2 + 1, Color.FromArgb(IronPalette[v].r, IronPalette[v].g, IronPalette[v].b));
                    }
                }

                // Queue Image for display
                bmpQueue.Enqueue(bmp);
                if (bmpQueue.Count > 1) bmpQueue.Dequeue(); 
            }

            y = 10;
            foreach(Bitmap b in bmpQueue.Reverse())
            {
                e.Graphics.DrawImage(b, 10, y);
                y += b.Height + 10;
            }
        }

        // Button to capture external reference or switch to internal shutter.
        private void button1_Click(object sender, EventArgs e)
        {
            captureReference = true;
        }

        // Button to toggle between automatic ranging or manual.
        private void button2_Click(object sender, EventArgs e)
        {
            rangeOverrideToggle = true;
        }

        // Not needed events.
        private void label2_Click(object sender, EventArgs e)
        {

        }

        private void label5_Click(object sender, EventArgs e)
        {

        }

        private void label6_Click(object sender, EventArgs e)
        {

        }

        private void label7_Click(object sender, EventArgs e)
        {

        }

        private void trackBar1_Scroll(object sender, EventArgs e)
        {

        }

        private void trackBar2_Scroll(object sender, EventArgs e)
        {

        }
    }
}
